#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <stdbool.h>
#include <zlib.h>

uint8_t wanted4017, *found4017;

bool errorOccurred = false;

void static inline put_lsb32(void *b, uint32_t x) {
	uint8_t* buffer = (uint8_t*) b;
	buffer[0] = x & 0xff;
	buffer[1] = x >>  8;
	buffer[2] = x >> 16;
	buffer[3] = x >> 24;
}
static inline uint32_t get_lsb32(uint8_t* buf) { return buf[0]      | (buf[1]<<8) | (buf[2]<<16) | (buf[3]<<24); }

bool processFile(char* fileName, uint8_t** _fileData, size_t* fileSize) {
	bool changesMade =false;
	uint8_t* fileData = *_fileData;
	if ((get_lsb32(&fileData[0x34]) +0x34) <0x88 || get_lsb32(&fileData[0x84])==0) {
		printf("%s: Does not use the NES APU\n", fileName);
		return false;
	}
	
	uint8_t* s =fileData + fileData[0x34] +0x34;
	uint8_t c;
	found4017 =NULL;
	while ((c =*s) != 0x66) {
		if (c==0xB4) {
			uint8_t reg = s[1];
			if (reg ==0x17) found4017 =s+2;
		}
		if (c==0x62 || c==0x63 || c==0x66 || (c>=0x70 && c<=0x8F)) s+=1; else
		if (c==0x4F || c==0x50) s+=2; else
		s+=3;
	}
	if (found4017) {
		printf("%s: $4017 value already set to %02X\n", fileName, *found4017);
	} else {
		printf("%s: Added $4017 value %02X to the beginning of the file\n", fileName, wanted4017);
		s =fileData + fileData[0x34] +0x34;

		uint32_t position = s-fileData;
		*fileSize += 3;
		fileData =realloc(fileData, *fileSize);
		*_fileData =fileData;
		s=fileData +position;
		
		memmove(s+3, s, *fileSize-3-position);
		*s++ =0xB4;
		*s++ =0x17;
		*s++ =wanted4017;
		
		// update total file size
		uint32_t rofsEOF = get_lsb32(&fileData[0x04]); if (rofsEOF) {
			rofsEOF += 3;
			put_lsb32(&fileData[0x04], rofsEOF);
		}
		// update GD3 offset
		uint32_t rofsGD3 = get_lsb32(&fileData[0x14]); if (rofsGD3) {
			rofsGD3 += 3;
			put_lsb32(&fileData[0x14], rofsGD3);
		}
		// update loop offset
		uint32_t rofsLoop = get_lsb32(&fileData[0x1C]); if (rofsLoop) {
			rofsLoop += 3;
			put_lsb32(&fileData[0x1C], rofsLoop);
		}
		changesMade =true;
	}
	return changesMade;
}

void showUsage(void) {
	printf("Usage: vgm4017 4017_value file(s)\n");
}

void loadAndProcessFile(char* fileName) {
	struct gzFile_s *fileHandle = gzopen(fileName, "rb");
	if (fileHandle) {
		/* Read the first eight bytes to check for the .VGM signature, and get the total file size */
		uint8_t firstEightBytes[8];
		gzread(fileHandle, firstEightBytes, 8);
		if (memcmp(firstEightBytes, "Vgm ", 4)==0) {
			size_t fileSize = get_lsb32(&firstEightBytes[4]) +4;
			uint8_t* fileData = malloc(fileSize);
			if (fileData) {
				gzrewind(fileHandle);
				size_t readSize = gzread(fileHandle, fileData, fileSize);
				if (readSize == fileSize) {
					if (processFile(fileName, &fileData, &fileSize)) {
						/* Close handle opened for read access, open handle for write access  (gzip interface does not allow both */
						gzclose(fileHandle);
						fileHandle =gzopen(fileName, "wb");
						if (fileHandle) {
							size_t wroteSize = gzwrite(fileHandle, fileData, fileSize);
							if (wroteSize != fileSize) {
								fprintf(stderr, "%s: Cannot write; wrote only %d bytes, wanted %d bytes\n", fileName, wroteSize, fileSize);
								errorOccurred = true;
							}
							printf("%s: File rewritten.\n", fileName);
						} else {
							fprintf(stderr, "%s: Could not open file for write access\n", fileName);
							errorOccurred = true;
						}
					} else {
						printf("%s: No changes made.\n", fileName);
					}
				} else {
					fprintf(stderr, "%s: Unexpected end of file; received only %d bytes, wanted %d bytes\n", fileName, readSize, fileSize);
					errorOccurred = true;
				}
				free(fileData);
			} else {
				fprintf(stderr, "%s: Cannot allocate %d bytes of memory\n", fileName, fileSize);
				errorOccurred = true;
			}
		} else {
			fprintf(stderr, "%s: Not a valid .VGM file\n", fileName);
			errorOccurred = true;
		}
		gzclose(fileHandle);
	} else {
		perror(fileName);
		errorOccurred = true;
	}
}

int main(int argc, char** argv) {
	if (argc <= 2) {
		showUsage();
		return EXIT_FAILURE;
	}
	char *afterValue;
	wanted4017 = strtol(argv[1], &afterValue, 16);
	if (*afterValue != '\0') {
		fprintf(stderr, "Invalid characters in 4017 value: %s\n", afterValue);
		return EXIT_FAILURE;
	}
	for (int arg =2; arg <argc; arg++) {
		loadAndProcessFile(argv[arg]);
	}
	return errorOccurred? EXIT_FAILURE: EXIT_SUCCESS;
}